/*
 * esp8266 wifi bridge for klipper
 *
 * author: Dikarev Evgeny
 * test: socat -dd pty,link=/home/klipper/printer_fb5_data/comms/virtualcom,ispeed=b250000,ospeed=b250000,raw,echo=0 TCP4-CONNECT:192.168.2.34:8888,nodelay,forever,interval=1
 */

#include <ESP8266WiFi.h>

#define ENABLE_DEBUG                    0

#define STA_SSID                        "my_ssid"
#define STA_PSK                         "ssid_password"

#define Def_BridgeServerPort            8888
#define Def_BridgeServerMaxClients      2
#define BAUD_SERIAL                     250000
#define RXBUFFERSIZE                    256

const char*     ssid                    = STA_SSID;
const char*     password                = STA_PSK;
const uint16_t  bridgeServerPort        = Def_BridgeServerPort;

// listening tcp server object
WiFiServer bridgeServer(bridgeServerPort);

// single client object
WiFiClient bridgeClient;

// object flag the presence of a connected client
bool bridgeClientExist = false;

// buffer size for receiving/transmitting data
const uint16_t bridgeClientBufferSize   = 256;

// buffer for receiving data from a tcp client
uint8_t  bridgeClientBuffer_from[bridgeClientBufferSize];
uint16_t bridgeClientBuffer_fromSize    = 0;

// buffer for receiving data to tcp client
uint8_t  bridgeClientBuffer_to[bridgeClientBufferSize];
uint16_t bridgeClientBuffer_toSize      = 0;

#if ENABLE_DEBUG
void printClientStatus();
void printClientIP();
#endif

void read_fromClientToBuffer();
void read_fromSerialToBuffer();

void write_fromBufferToSerial();
void write_fromBufferToClient();


void setup()
{
    // init serial
    Serial.begin(BAUD_SERIAL);
    Serial.setRxBufferSize(RXBUFFERSIZE);

#if ENABLE_DEBUG
    Serial.println();
    Serial.printf("-- init begin --");
    Serial.println();

    Serial.print("Connecting to ");
    Serial.print(ssid);
    Serial.println();
#endif

  /*
   * Explicitly set the ESP8266 as a WiFi client, otherwise it will default
   * to acting as both a client and an access point, which can cause network problems
   * with your other WiFi devices on your WiFi network 
   */
    WiFi.setSleepMode(WIFI_NONE_SLEEP);
    WiFi.setPhyMode(WIFI_PHY_MODE_11N);
    WiFi.mode(WIFI_STA);
    WiFi.begin(ssid, password);

    // establish a wifi connection to the server
    while (WiFi.status() != WL_CONNECTED)
    {
        delay(100);
#if ENABLE_DEBUG
        Serial.print(".");
#endif
    }

#if ENABLE_DEBUG
    Serial.println();
    Serial.println("WiFi connected");
    Serial.print("IP address: ");
    Serial.println(WiFi.localIP());
#endif

    // start bridge server
    bridgeServer.begin(bridgeServerPort, Def_BridgeServerMaxClients);
    bridgeServer.setNoDelay(true);

    // Initialize the LED_BUILTIN pin as an output
    pinMode(LED_BUILTIN, OUTPUT);

    // signal that we are loaded
    digitalWrite(LED_BUILTIN, HIGH); delay(200); digitalWrite(LED_BUILTIN, LOW); delay(200);
    digitalWrite(LED_BUILTIN, HIGH); delay(200); digitalWrite(LED_BUILTIN, LOW); delay(200);
    digitalWrite(LED_BUILTIN, HIGH); delay(200); digitalWrite(LED_BUILTIN, LOW); delay(200);
    digitalWrite(LED_BUILTIN, HIGH);

#if ENABLE_DEBUG
    Serial.printf("-- init complete --");
    Serial.println();
#endif
}

// the loop function runs over and over again forever
void loop()
{
    // check the connection with the current client
    // i.e. if there is a client, but there is no connection with it,
    //  then we close / stop the client.
    if(bridgeClientExist && !bridgeClient.connected())
    {
#if ENABLE_DEBUG
        Serial.print("disconnect: ");
        printClientIP();
        Serial.println();
#endif
        bridgeClient.stop();
        bridgeClientExist = false;
    }


    // check for new connections
    if (bridgeServer.hasClient())
    {
        // if there are no clients yet, then we create a connection
        if(!bridgeClientExist)
        {
            bridgeClient = bridgeServer.accept();
            bridgeClientExist = true;
#if ENABLE_DEBUG
            Serial.print("new connection ");
            printClientIP();
            Serial.println();
            printClientStatus();
#endif
        }
        else
        {
            Serial.println("reject connection");
            bridgeServer.accept().stop();
        }
    }

    // check for the presence of a connected client,
    // if there are still no clients, then we are waiting for them
    if(!bridgeClientExist)
    {
        return;
    }

    // check for a connected client
    if(!bridgeClient.connected())
    {
        return;
    }

    // we check the presence of input data from the tcp client.
    // subtract the input
    read_fromClientToBuffer();

    // if the serial port is ready for writing,
    // then we send read data to it
    write_fromBufferToSerial();

    // check for data on the serial port.
    // if the serial port is ready for reading (there is data),
    // then we subtract them
    read_fromSerialToBuffer();


    // if the client is ready to transfer data, then transfer this data
    write_fromBufferToClient();

}


/************************************************************************/
void printClientIP()
{
#if ENABLE_DEBUG
    Serial.print(bridgeClient.remoteIP()[0]);
    Serial.print(".");
    Serial.print(bridgeClient.remoteIP()[1]);
    Serial.print(".");
    Serial.print(bridgeClient.remoteIP()[2]);
    Serial.print(".");
    Serial.print(bridgeClient.remoteIP()[3]);
    Serial.print(":");
    Serial.print(bridgeClient.remotePort());
#endif
}

void printClientStatus()
{
#if ENABLE_DEBUG
    Serial.print("client: ");
    if(bridgeClient.connected())
    {
        Serial.print("connected");
    }
    else
    {
        Serial.print("not connected");
    }

    Serial.print("; status = ");
    Serial.print(bridgeClient.status());
    Serial.println();
#endif
}

void read_fromClientToBuffer()
{
    int availableBytes = bridgeClient.available();
    if(availableBytes == 0)
    {
        return;
    }

    // calculate how much space is left in the buffer
    int bridgeClientBuffer_leftSize = bridgeClientBufferSize - bridgeClientBuffer_fromSize;

    // if there is nothing left - exit
    if(bridgeClientBuffer_leftSize == 0)
    {
        return;
    }

    // read the input
    int readedBytes = bridgeClient.read(
        bridgeClientBuffer_from + bridgeClientBuffer_fromSize
        , std::min(availableBytes, bridgeClientBuffer_leftSize));

#if ENABLE_DEBUG
    Serial.println("read4Client:");
    Serial.write(bridgeClientBuffer_from + bridgeClientBuffer_fromSize, readedBytes);
    Serial.println("--- end ---");
    Serial.println();
#endif

    bridgeClientBuffer_fromSize += readedBytes;

#if ENABLE_DEBUG
        Serial.print("bridgeClientBuffer_fromSize: ");
        Serial.println(bridgeClientBuffer_fromSize);

        Serial.print("bridgeClientBuffer_toSize: ");
        Serial.println(bridgeClientBuffer_toSize);
#endif
    if((bridgeClientBuffer_fromSize + bridgeClientBuffer_toSize) > 0)
    {
        // turn on the indication of the presence of data in the buffer
        digitalWrite(LED_BUILTIN, HIGH);

#if ENABLE_DEBUG
        Serial.println("led on");
#endif
    }
}

void read_fromSerialToBuffer()
{
    int availableBytes = Serial.available();
    if(availableBytes == 0)
    {
        return;
    }

    // calculate how much space is left in the buffer
    int bridgeClientBuffer_leftSize = bridgeClientBufferSize - bridgeClientBuffer_toSize;

    // if there is nothing left - exit
    if(bridgeClientBuffer_leftSize == 0)
    {
        return;
    }

    // read the input
    int readedBytes = Serial.read(
        bridgeClientBuffer_to + bridgeClientBuffer_toSize
        , std::min(availableBytes, bridgeClientBuffer_leftSize));

#if ENABLE_DEBUG
    Serial.println("read4Serial:");
    Serial.write(bridgeClientBuffer_to + bridgeClientBuffer_toSize, readedBytes);
    Serial.println("--- end ---");
    Serial.println();
#endif

    bridgeClientBuffer_toSize += readedBytes;
}

void write_fromBufferToSerial()
{
    // check the presence of data in the buffer from the tcp client
    if(bridgeClientBuffer_fromSize == 0)
    {
        return;
    }

    // if the serial port is not ready for recording - skip recording
    if(!Serial.availableForWrite())
    {
        return;
    }

#if ENABLE_DEBUG
    Serial.println("write2Serial:");
#endif

    // if the serial port is ready for writing, then we send read data to it
    int writedBytes = Serial.write(bridgeClientBuffer_from, bridgeClientBuffer_fromSize);
    memcpy(bridgeClientBuffer_from, bridgeClientBuffer_from + writedBytes, writedBytes);
    bridgeClientBuffer_fromSize -= writedBytes;

#if ENABLE_DEBUG
    Serial.println("--- end ---");
    Serial.println();
#endif
}

void write_fromBufferToClient()
{
    // check the presence of data in the buffer from the tcp client
    if(bridgeClientBuffer_toSize == 0)
    {
        return;
    }

    // if tcp client is not ready for recording - skip recording
    if(!bridgeClient.availableForWrite())
    {
        return;
    }

    // send the read data from the buffer to the socket
    int writedBytes = bridgeClient.write(bridgeClientBuffer_to, bridgeClientBuffer_toSize);

#if ENABLE_DEBUG
    Serial.println("write2Client:");
    Serial.write(bridgeClientBuffer_to, writedBytes);
    Serial.println("--- end ---");
    Serial.println();
#endif

    memcpy(bridgeClientBuffer_to, bridgeClientBuffer_to + writedBytes, writedBytes);
    bridgeClientBuffer_toSize -= writedBytes;

#if ENABLE_DEBUG
    Serial.print("bridgeClientBuffer_fromSize: ");
    Serial.println(bridgeClientBuffer_fromSize);

    Serial.print("bridgeClientBuffer_toSize: ");
    Serial.println(bridgeClientBuffer_toSize);
#endif
    if( (bridgeClientBuffer_fromSize + bridgeClientBuffer_toSize) == 0)
    {
        // turn off the indication of the presence of data in the buffer
        digitalWrite(LED_BUILTIN, LOW);
#if ENABLE_DEBUG
        Serial.println("led off");
#endif
    }
}
